//	Entity.cpp - Class representing one of the game's entities.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "helper/ColourHelpers.h"
#include "helper/MiscHelper.h"
#include "drawing/Drawer.h"
#include "WakeUpSounder.h"
#include "Entity.h"
#include "Random.h"

//------------------------------------------------------------------------------
Entity::Entity(const TwoFloats& pos,
			   const ThreeFloats& col,
			   const std::wstring& im,
			   float rad,
			   const TwoFloats& drawSize,
			   shared_ptr<WakeUpSounder>& sound,
			   float dist):
position(pos),
colour(col),
image(im),
radius(rad),
rotDir((Random::getGlobal().getRandomFloat() < 0.5f) ? -1.0f : 1.0f),
direction(-1.0f, 0.0f),
dirTween(direction, direction, 0.02f),
sounder(sound)
{
	if(pos.x < 0.0f)
	{
		direction.x = 0.5f + (Random::getGlobal().getRandomFloat() * 0.5f);
		direction.y = Random::getGlobal().getRandomFloat() - 0.5f;
		direction = direction.normal();

		dirTween.reset(TwoFloats(Random::getGlobal().getRandomFloat(),
								 Random::getGlobal().getRandomFloat() - 0.5f));
	}
	else if(pos.x > drawSize.x)
	{
		direction.x = -(0.5f + (Random::getGlobal().getRandomFloat() * 0.5f));
		direction.y = Random::getGlobal().getRandomFloat() - 0.5f;
		direction = direction.normal();

		dirTween.reset(TwoFloats(-(Random::getGlobal().getRandomFloat()),
								 Random::getGlobal().getRandomFloat() - 0.5f));
	}
	else if(pos.y < 0.0f)
	{
		direction.x = Random::getGlobal().getRandomFloat() - 0.5f;
		direction.y = 0.5f + (Random::getGlobal().getRandomFloat() * 0.5f);
		direction = direction.normal();

		dirTween.reset(TwoFloats(Random::getGlobal().getRandomFloat() - 0.5f,
								 Random::getGlobal().getRandomFloat()));
	}
	else if(pos.y > drawSize.y)
	{
		direction.x = Random::getGlobal().getRandomFloat() - 0.5f;
		direction.y = -(0.5f + (Random::getGlobal().getRandomFloat() * 0.5f));
		direction = direction.normal();

		dirTween.reset(TwoFloats(Random::getGlobal().getRandomFloat() - 0.5f,
								 -(Random::getGlobal().getRandomFloat())));
	}

	droneIndex = sounder->startDrone(ColourHelpers::rgbToHsl(colour).x, (position.x/drawSize.x));
	sounder->setDroneDist(droneIndex, dist);
}

//------------------------------------------------------------------------------
Entity::~Entity()
{
	sounder->stopDrone(droneIndex);
}

//------------------------------------------------------------------------------
void Entity::update(const TwoFloats& playerPos,
					InteractFunc func,
					const TwoFloats& drawSize)
{
	TwoFloats wanderPos(position + (direction * 1.0f));
	TwoFloats funcPos(position);
	float dist(position.getDistance(playerPos));
	TwoFloats newPos;

	if(dist < 256.0f)
	{
		func(funcPos, playerPos, rotDir);

		newPos = wanderPos + ((funcPos - wanderPos) * fakeLog(1.0f - (dist/256.0f)));
	}
	else
		newPos = wanderPos;

	if(newPos.x < -32.0f)
		newPos.x = -32.0f;
	else if(newPos.x > (drawSize.x + 32.0f))
		newPos.x = (drawSize.x + 32.0f);

	if(newPos.y < -32.0f)
		newPos.y = -32.0f;
	else if(newPos.y > (drawSize.y + 32.0f))
		newPos.y = (drawSize.y + 32.0f);

	speed = newPos.getDistance(position);
	position = newPos;

	dirTween.tick();
	if(dirTween.getFinished())
	{
		TwoFloats newDir((Random::getGlobal().getRandomFloat() * 2.0f) - 1.0f,
						 (Random::getGlobal().getRandomFloat() * 2.0f) - 1.0f);

		if(position.x <= 128.0f)
			newDir.x = 0.5f + (Random::getGlobal().getRandomFloat() * 0.5f);
		else if(position.x >= (drawSize.x - 128.0f))
			newDir.x = -(0.5f + (Random::getGlobal().getRandomFloat() * 0.5f));
		if(position.y <= 128.0f)
			newDir.y = 0.5f + (Random::getGlobal().getRandomFloat() * 0.5f);
		else if(position.y >= (drawSize.y - 128.0f))
			newDir.y = -(0.5f + (Random::getGlobal().getRandomFloat() * 0.5f));

		dirTween.setDelta(0.01f + (Random::getGlobal().getRandomFloat() * 0.02f));
		dirTween.reset(newDir.normal());
	}

	sounder->setDroneAmp(droneIndex, clamp<float>((speed*0.1f), 0.0f, 1.0f));
	sounder->setDronePan(droneIndex, (position.x/drawSize.x));
}

//------------------------------------------------------------------------------
void Entity::draw(shared_ptr<Drawer>& d)
{
	d->drawImage(image,
				 position - radius,
				 1.0f,
				 colour);
}
