//	Entity.h - Class representing one of the game's entities.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef ENTITY_H_
#define ENTITY_H_

#include "ThreeFloats.h"
#include "c0xHeaders.h"
#include "TwoFloats.h"
#include "Tweener.h"

class WakeUpSounder;
class Drawer;

///	This is the function by which the player interacts with the Entity.
/*!
	Parameters are: Entity position; player position; rotate direction.
 */
typedef void (*InteractFunc)(TwoFloats&, const TwoFloats&, float);

///	Class representing one of the game's entities.
class Entity
{
  public:
	///	Constructor.
	Entity(const TwoFloats& pos,
		   const ThreeFloats& col,
		   const std::wstring& im,
		   float rad,
		   const TwoFloats& drawSize,
		   shared_ptr<WakeUpSounder>& sound,
		   float dist);
	///	Destructor.
	~Entity();

	///	Moves the entity.
	void update(const TwoFloats& playerPos,
				InteractFunc func,
				const TwoFloats& drawSize);
	///	Draws the entity.
	void draw(shared_ptr<Drawer>& d);

	///	Returns the entity's position.
	const TwoFloats& getPosition() const {return position;};
	///	Returns the entity's radius.
	float getRadius() const {return radius;};
	///	Returns the entity's image.
	const std::wstring& getImage() const {return image;};
	///	Returns the entity's colour.
	const ThreeFloats& getColour() const {return colour;};
  private:
	///	The entity's position.
	TwoFloats position;
	///	The entity's colour.
	ThreeFloats colour;
	///	The entity's image.
	std::wstring image;
	///	The entity's radius.
	float radius;
	///	The entity's rotation direction.
	float rotDir;

	///	Used to wander the entity.
	TwoFloats direction;
	///	Used to wander the entity.
	SmoothstepTweener<TwoFloats> dirTween;

	///	Our sine drone index.
	int droneIndex;
	///	Our speed (determines our drone amp).
	float speed;

	///	Our copy of the sounder.
	shared_ptr<WakeUpSounder>& sounder;
};

#endif
