//	GamePlayState.cpp - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "drawing/FancyImage.h"
#include "WakeUpGlobalData.h"
#include "drawing/Drawer.h"
#include "GamePlayState.h"
#include "ColourHelpers.h"
#include "WakeUpSounder.h"
#include "Application.h"
#include "MiscHelper.h"
#include "Explosion.h"
#include "Entity.h"
#include "Random.h"

#include <sstream>
#include <iomanip>

//------------------------------------------------------------------------------
GameStateCreator<GamePlayState> GamePlayState::creator;

//------------------------------------------------------------------------------
GamePlayState::GamePlayState(Application& app, SounderPtr& sound):
GameState(app, sound),
mousePos(640.0f, 360.0f),
filteredPos(640.0f, 360.0f),
imageIndex(0),
entityCount(180),
timeAttack(true),
timer(180.0f),
timerWait(true),
endCount(60),
shakeAmp(0.0f),
shakeIndex(0.0f),
firstGo(true),
sounder(dynamic_pointer_cast<WakeUpSounder>(sound))
{
	app.setMousePos(mousePos);

	if(dynamic_pointer_cast<WakeUpGlobalData>(app.getGlobalData())->getPlayground())
		timeAttack = false;

	for(int i=0;i<4;++i)
		keyIsDown[i] = false;

	addEntity();
}

//------------------------------------------------------------------------------
GamePlayState::~GamePlayState()
{

}

//------------------------------------------------------------------------------
void GamePlayState::update()
{
	float mag(joyPos.getMagnitude());
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());
	InteractFunc functions[] = {repelFunc,
								pauseFunc,
								rotateFunc};

	if(timeAttack && !timerWait && (timer > 0.0f))
	{
		uint32_t ticks(getApplication().getTimeSinceStart());

		timer -= (float)(ticks-lastTicks)/1000.0f;
		lastTicks = ticks;
	}
	else if(timeAttack && (timer <= 0.0f) && (endCount > 0))
		--endCount;

	if(!timeAttack || (!timerWait && (timer > 0.0f)))
	{
		if(entities.size() < 32)
		{
			--entityCount;
			if(!entityCount)
			{
				addEntity();
				entityCount = 180;
			}
		}

		if(mag > 0.5f)
		{
			TwoFloats tempJoy(joyPos);
			TwoFloats tempPos(mousePos);

			if(mag < 1.0f)
			{
				float scale((mag - 0.5f)/0.5f);

				tempJoy *= fakeExp(scale);
			}
			tempPos += tempJoy;

			if(tempPos.x < 0.0f)
				tempPos.x = 0.0f;
			else if(tempPos.x >= drawSize.x)
				tempPos.x = drawSize.x;

			if(tempPos.y < 0.0f)
				tempPos.y = 0.0f;
			else if(tempPos.y >= drawSize.y)
				tempPos.y = drawSize.y;

			mousePos = tempPos;
		}

		TwoFloats keyPos;
		if(keyIsDown[0] && !keyIsDown[2])
			keyPos.x = -1.0f;
		else if(!keyIsDown[0] && keyIsDown[2])
			keyPos.x = 1.0f;
		if(keyIsDown[1] && !keyIsDown[3])
			keyPos.y = -1.0f;
		else if(!keyIsDown[1] && keyIsDown[3])
			keyPos.y = 1.0f;
		if(keyPos.getMagnitude() > 0.0f)
		{
			mousePos += keyPos * 3.0f;
		}

		if(filteredPos.getDistanceSquared(mousePos) < (0.25f * 0.25f))
			filteredPos = mousePos;
		else
		{
			TwoFloats tempPos(filteredPos);

			//tempPos += (mousePos - tempPos) * 0.033f;
			tempPos += (mousePos - tempPos) * 0.2f;
			/*if((tempPos - filteredPos).getMagnitude() > 2.0f)
				tempPos = filteredPos + ((tempPos - filteredPos).normal() * 2.0f);*/

			filteredPos = tempPos;
		}

		if(mousePos.getDistance(filteredPos) > 4.0f)
		{
			TwoFloats vec((mousePos - filteredPos).normal());

			mousePos = filteredPos + (vec * 4.0f);
		}

		for(unsigned int i=0;i<entities.size();++i)
		{
			TwoFloats effectorPos(getNearestEffector(entities[i]->getPosition()));

			entities[i]->update(effectorPos, functions[imageIndex], drawSize);
		}
		for(unsigned int i=0;i<triangles.size();++i)
			triangles[i]->update(filteredPos, functions[imageIndex], drawSize);

		collisions();
	}

	for(int i=(plusOnes.size()-1);i>=0;--i)
	{
		plusOnes[i]->update();
		if(plusOnes[i]->getFinished())
			plusOnes.erase(plusOnes.begin() + i);
	}

	for(int i=(explosions.size()-1);i>=0;--i)
	{
		explosions[i]->update();
		if(explosions[i]->getDead())
			explosions.erase(explosions.begin() + i);
	}

	shakeIndex += 0.3f;
	if(shakeIndex > (2.0f * NIALL_PI))
		shakeIndex -= 2.0f * NIALL_PI;
	if(shakeAmp > 0.0f)
		shakeAmp -= 0.03f;
}

//------------------------------------------------------------------------------
void GamePlayState::draw(DrawerPtr& d, float timestepInterp)
{
	float shakeVal(0.0f);
	const std::wstring back[] = {L"BlueBack",
								 L"GreenBack",
								 L"RedBack"};
	const std::wstring front[] = {L"BlueFront",
								  L"GreenFront",
								  L"RedFront"};
	const TwoFloats& drawSize(d->getDrawingAreaSize());

	d->fillRect(0.0f, d->getDrawingAreaSize(), 1.0f);

	if(shakeAmp > 0.0f)
		shakeVal = ((sinf(shakeIndex) * 0.01f) + 0.01f) * shakeAmp;
	d->drawImage(d->getFrameBufferImage(),
				 TwoFloats(-(shakeVal * drawSize.x * 0.5f), 0.0f),
				 TwoFloats(1.0f + shakeVal, 1.0f));

	d->startFrameBufferDrawing();

	if(firstGo)
	{
		d->fillRect(0.0f, drawSize, 1.0f);
		firstGo = false;
	}

	for(unsigned int i=0;i<entities.size();++i)
		entities[i]->draw(d);
	for(unsigned int i=0;i<triangles.size();++i)
		triangles[i]->draw(d);

	d->drawImage(front[imageIndex], filteredPos - 45.0f);

	update();

	for(unsigned int i=0;i<entities.size();++i)
		entities[i]->draw(d);
	for(unsigned int i=0;i<triangles.size();++i)
		triangles[i]->draw(d);

	d->drawImage(front[imageIndex], filteredPos - 45.0f);

	d->endFrameBufferDrawing();

	for(unsigned int i=0;i<explosions.size();++i)
		explosions[i]->draw(d);

	//d->drawImage(back[imageIndex], filteredPos - 45.0f);
	//d->drawImage(front[imageIndex], filteredPos - 45.0f);

	for(unsigned int i=0;i<plusOnes.size();++i)
		plusOnes[i]->draw(d);

	if(timeAttack)
	{
		//Draw timer.
		{
			std::wstringstream tempstr;

			tempstr << (int)(ceilf(timer)/60.0f);
			tempstr << L":" << std::setw(2) << std::setfill(L'0') << ((int)ceilf(timer) % 60);
			d->drawText(tempstr.str(),
						L"BigFont",
						TwoFloats(4.0f, 80.0f + 4.0f),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						0.0f,
						0.5f);
			d->drawText(tempstr.str(),
						L"BigFont",
						TwoFloats(0.0f, 80.0f),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						1.0f);
		}

		//Draw score.
		if(timer > 0.0f)
		{
			std::wstringstream tempstr;

			tempstr << counter.getScore();
			d->drawText(tempstr.str(),
						L"BigFont",
						TwoFloats(4.0f, (drawSize.y - 32.0f) + 4.0f),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						0.0f,
						0.5f);
			d->drawText(tempstr.str(),
						L"BigFont",
						TwoFloats(0.0f, (drawSize.y - 32.0f)),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						1.0f);
		}
		else
		{
			std::wstringstream tempstr;

			tempstr << counter.getScore();
			d->drawText(tempstr.str(),
						L"HugeFont",
						TwoFloats(4.0f, (drawSize.y * 0.5f) + 24.0f + 4.0f),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						0.0f,
						0.5f);
			d->drawText(tempstr.str(),
						L"HugeFont",
						TwoFloats(0.0f, (drawSize.y * 0.5f) + 24.0f),
						drawSize.x,
						TextLayout::Centre,
						1.0f,
						1.0f);
		}
	}
}

//------------------------------------------------------------------------------
void GamePlayState::keyDown(const std::wstring& action,
							const std::wstring& keyPressed,
							int keyModifier)
{
	if(timeAttack && timerWait && !firstGo)
	{
		timerWait = false;
		lastTicks = getApplication().getTimeSinceStart();
	}

	if(timeAttack && !endCount)
		nextState(L"TitleScreen");
	else if(action == L"Escape")
		nextState(L"TitleScreen");
	else if(keyPressed == L"f5")
		nextState(L"GamePlayState");
	else if(action == L"P1Blue")
	{
		imageIndex = 0;
		sounder->triggerBlueSynth();
	}
	else if(action == L"P1Green")
	{
		imageIndex = 1;
		sounder->triggerGreenSynth();
	}
	else if(action == L"P1Red")
	{
		imageIndex = 2;
		sounder->triggerRedSynth();
	}
	else if(action == L"P1Left")
		keyIsDown[0] = true;
	else if(action == L"P1Up")
		keyIsDown[1] = true;
	else if(action == L"P1Right")
		keyIsDown[2] = true;
	else if(action == L"P1Down")
		keyIsDown[3] = true;
}

//------------------------------------------------------------------------------
void GamePlayState::keyUp(const std::wstring& action,
						  const std::wstring& keyPressed,
						  int keyModifier)
{
	if(action == L"P1Left")
		keyIsDown[0] = false;
	else if(action == L"P1Up")
		keyIsDown[1] = false;
	else if(action == L"P1Right")
		keyIsDown[2] = false;
	else if(action == L"P1Down")
		keyIsDown[3] = false;
}

//------------------------------------------------------------------------------
void GamePlayState::mouseMove(const TwoFloats& pos)
{
	//mousePos = pos;
}

//------------------------------------------------------------------------------
void GamePlayState::joyMove(const std::wstring& action,
							float value,
							int joystick,
							int axis)
{
	if(timeAttack && timerWait)
	{
		timerWait = false;
		lastTicks = getApplication().getTimeSinceStart();
	}

	if((action == L"P1Up") || (action == L"P1Down"))
		joyPos.y = value * 3.0f;
	else if((action == L"P1Left") || (action == L"P1Right"))
		joyPos.x = value * 3.0f;
}

//------------------------------------------------------------------------------
void GamePlayState::repelFunc(TwoFloats& pos,
							  const TwoFloats& playerPos,
							  float rotDir)
{
	TwoFloats newVal(pos - ((playerPos -  pos) * 0.1f));

	if(newVal.getMagnitude() > 3.0f)
		newVal = pos + ((newVal - pos).normal() * 3.0f);

	pos = newVal;
}

//------------------------------------------------------------------------------
void GamePlayState::attractFunc(TwoFloats& pos,
							    const TwoFloats& playerPos,
							    float rotDir)
{
	TwoFloats newVal(pos + ((playerPos -  pos) * 0.05f));

	if(newVal.getMagnitude() > 3.0f)
		newVal = pos + ((newVal - pos).normal() * 3.0f);

	float dist(pos.getDistance(playerPos));
	if(dist > 128.0f)
		pos = newVal;
	else if(dist > 32.0f)
		pos += (newVal - pos) * fakeLog((dist-45.0f)/(128.0f-45.0f));
}

//------------------------------------------------------------------------------
void GamePlayState::rotateFunc(TwoFloats& pos,
							   const TwoFloats& playerPos,
							   float rotDir)
{
	float index(playerPos.getAngle(pos));
	float rad(pos.getDistance(playerPos));

	index += rotDir * 0.0266f;

	pos.x = playerPos.x + (sinf(index) * rad);
	pos.y = playerPos.y + (cosf(index) * rad);
}

//------------------------------------------------------------------------------
void GamePlayState::pauseFunc(TwoFloats& pos,
							  const TwoFloats& playerPos,
							  float rotDir)
{
	
}

//------------------------------------------------------------------------------
void GamePlayState::addEntity()
{
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());
	TwoFloats starts[] = {TwoFloats(-32.0f, 0.0f),
						  TwoFloats(0.0f, -32.0f),
						  TwoFloats(drawSize.x + 32.0f, 0.0f),
						  TwoFloats(0.0f, drawSize.y + 32.0f)};
	int index(Random::getGlobal().getRandomInt(4));

	TwoFloats pos(starts[index]);

	if(pos.x == 0.0f)
		pos.x = Random::getGlobal().getRandomFloat() * drawSize.x;
	else if(pos.y == 0.0f)
		pos.y = Random::getGlobal().getRandomFloat() * drawSize.y;

	ThreeFloats col(Random::getGlobal().getRandomFloat(), 1.0f, 0.8f);
	//ThreeFloats col(Random::getGlobal().getRandomFloat(), 1.0f, 0.5f);

	entities.push_back(shared_ptr<Entity>(new Entity(pos,
										  ColourHelpers::hslToRgb(col),
										  L"Entity",
										  28.0f,
										  drawSize,
										  sounder,
										  0.0f)));
	//sounder->triggerNewEntitySynth(col.x);
}

//------------------------------------------------------------------------------
void GamePlayState::collisions()
{
	const TwoFloats& drawSize(getApplication().getDrawingAreaSize());

	for(int i=(entities.size()-1);i>=0;--i)
	{
		const TwoFloats& pos1(entities[i]->getPosition());
		bool collided(false);

		for(int j=(i-1);j>=0;--j)
		{
			if(entities[i]->getImage() == entities[j]->getImage())
			{
				const TwoFloats& pos2(entities[j]->getPosition());
				float radSq(entities[i]->getRadius() - 6.0f);

				radSq *= 2.0f;
				radSq *= radSq;

				if(pos1.getDistanceSquared(pos2) < radSq)
				{
					int index(getNextImage(entities[i]->getImage()));

					if(index < 4)
					{
						TwoFloats pos((pos1 + pos2) * 0.5f);
						ThreeFloats col1(ColourHelpers::rgbToHsl(entities[i]->getColour()));
						ThreeFloats col2(ColourHelpers::rgbToHsl(entities[j]->getColour()));
						float r1(col1.x);
						float r2(col2.x);

						if(fabsf(r2 - r1) > 0.5f)
						{
							if(r2 > r1)
								r2 -= 1.0f;
							else if(r1 > r2)
								r1 -= 1.0f;
						}
						r1 = (r1 + r2) * 0.5f;
						if(r1 < 0.0f)
							r1 += 1.0f;
						ThreeFloats col(ColourHelpers::hslToRgb(ThreeFloats(r1, 1.0f, 0.8f)));
						//ThreeFloats col(ColourHelpers::hslToRgb(ThreeFloats(r1, 1.0f, 0.5f)));
						const std::wstring ims[] = {L"Entity",
												    L"Squidge",
												    L"Hexagon",
												    L"TriangleThing"};
						float rads[] = {28.0f,
										35.0f,
										40.0f,
										56.0f};

						entities.erase(entities.begin()+i);
						entities.erase(entities.begin()+j);

						if(index < 3)
						{
							entities.push_back(shared_ptr<Entity>(new Entity(pos,
																			 col,
																			 ims[index],
																			 rads[index],
																			 drawSize,
																			 sounder,
																			 (float)index/3.0f)));
							sounder->triggerNewEntitySynth(col.x);
						}
						else
						{
							triangles.push_back(shared_ptr<Entity>(new Entity(pos,
																			  col,
																			  ims[index],
																			  rads[index],
																			  drawSize,
																			  sounder,
																			  1.0f)));
							sounder->triggerNewTriangleSynth(col.x);

							//Add a new entity for every triangle created.
							addEntity();

							if(timeAttack)
							{
								//Add the '+1' popup.
								shared_ptr<FancyImage> im(new FancyImage(L"PlusOne",
																		 TwoFloats(pos.x - 35.5f, pos.y - 106.0f)));

								im->tweenPosition(TwoFloats(pos.x - 35.5f, pos.y - 192.0f), 0.01f);
								im->tweenAlpha(0.0f, 0.01f, FancyImage::Exponential, 0.2f);

								plusOnes.push_back(im);
								counter.incrementCount(L"score");
							}
						}

						explosions.push_back(shared_ptr<Explosion>(new Explosion(pos, col)));
						shakeAmp = 1.0f;

						collided = true;
						break;
					}
				}
			}
		}
		if(collided)
			break;
	}
}

//------------------------------------------------------------------------------
int GamePlayState::getNextImage(const std::wstring& im)
{
	const std::wstring images[] = {L"Entity",
								   L"Squidge",
								   L"Hexagon",
								   L"TriangleThing"};
	int i;

	for(i=0;i<4;++i)
	{
		if(im == images[i])
		{
			++i;
			break;
		}
	}
	
	return i;
}

//------------------------------------------------------------------------------
TwoFloats GamePlayState::getNearestEffector(const TwoFloats& pos)
{
	TwoFloats retval;
	float dist(100000.0f * 100000.0f);

	for(unsigned int i=0;i<triangles.size();++i)
	{
		float d(pos.getDistanceSquared(triangles[i]->getPosition()));

		if(d < dist)
		{
			dist = d;
			retval = triangles[i]->getPosition();
		}
	}

	float d(pos.getDistanceSquared(filteredPos));

	if(d < dist)
		retval = filteredPos;

	return retval;
}
