//	GamePlayState.h - The main gameplay state.
//  ----------------------------------------------------------------------------
//	This file is part of 'NiallsAVLib', base code for any kind of audiovisual
//	apps.
//	Copyright (C) 2012  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef GAMEPLAYSTATE_H_
#define GAMEPLAYSTATE_H_

#include "GameCounter.h"
#include "GameState.h"

#include <vector>

class WakeUpSounder;
class FancyImage;
class Explosion;
class Entity;

///	The main gameplay state.
class GamePlayState : public GameState
{
  public:
	///	Constructor.
	GamePlayState(Application& app, SounderPtr& sound);
	///	Destructor.
	~GamePlayState();

	///	Called frequently to update the state.
	void update();
	///	Called to draw the state.
	void draw(DrawerPtr& d, float timestepInterp);

	///	Called when a key down event is received.
	void keyDown(const std::wstring& action,
				 const std::wstring& keyPressed,
				 int keyModifier);
	///	Called when a key up event is received.
	void keyUp(const std::wstring& action,
			   const std::wstring& keyPressed,
			   int keyModifier);

	///	Used to test connected line drawing.
	void mouseMove(const TwoFloats& pos);
	///	Used to move the thing.
	void joyMove(const std::wstring& action,
				 float value,
				 int joystick,
				 int axis);

	///	The Repel function.
	static void repelFunc(TwoFloats& pos,
						  const TwoFloats& playerPos,
						  float rotDir);
	///	The Attract function.
	static void attractFunc(TwoFloats& pos,
							const TwoFloats& playerPos,
							float rotDir);
	///	The Rotate function.
	static void rotateFunc(TwoFloats& pos,
						   const TwoFloats& playerPos,
						   float rotDir);
	///	The Rotate function.
	static void pauseFunc(TwoFloats& pos,
						  const TwoFloats& playerPos,
						  float rotDir);

	///	Returns the name of this state.
	const std::wstring getName() const {return L"GamePlayState";};
	///	Used to construct instances of this state.
	static GameStateCreator<GamePlayState> creator;
  private:
	///	Helper method. Adds a random entity.
	void addEntity();
	///	Helper method. Handles Entity-Entity collisions.
	void collisions();
	///	Helper method. Returns the next image up for the passed-in image.
	int getNextImage(const std::wstring& im);
	///	Helper method. Returns the position of the nearest TriangleThing or Player.
	TwoFloats getNearestEffector(const TwoFloats& pos);

	///	Test.
	TwoFloats mousePos;
	///	Test.
	TwoFloats filteredPos;
	///	Test.
	int imageIndex;
	///	Test.
	TwoFloats joyPos;

	///	Used to enable keyboard controls.
	bool keyIsDown[4];

	///	The entities in the game.
	std::vector<shared_ptr<Entity> > entities;
	///	The TriangleThing entities in the game.
	std::vector<shared_ptr<Entity> > triangles;
	///	Used to add entities.
	int entityCount;

	///	Any active explosions.
	std::vector<shared_ptr<Explosion> > explosions;

	///	True if we're playing time attack mode.
	bool timeAttack;
	///	Used by the timer.
	uint32_t lastTicks;
	///	The timer.
	float timer;
	///	Used to wait till the player hits a button/joystick before the timer starts.
	bool timerWait;
	///	Used to wait a wee while after the game ends before we can return to the title screen.
	int endCount;

	///	The score counter.
	GameCounter counter;
	///	Used to display score stuff for the time attack mode.
	std::vector<shared_ptr<FancyImage> > plusOnes;

	///	Used to 'shake' the screen.
	float shakeAmp;
	///	Used to 'shake' the screen.
	float shakeIndex;

	///	Used to blank the background when the state starts.
	bool firstGo;

	///	Our sounder.
	shared_ptr<WakeUpSounder> sounder;
};

#endif
