//	WakeUpSounder.cpp - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'SoManyJaggedShards', an entry for 7DRL 2013.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "WakeUpSounder.h"
#include "MiscHelper.h"
#include "FilePath.h"
#include "UTF8File.h"
#include "Random.h"

#include "NewTriangle.inl"
#include "GreenSound.inl"
#include "BlueSound.inl"
#include "NewEntity.inl"
#include "RedSound.inl"

#include <cassert>
#include <sstream>
#include <string>

//------------------------------------------------------------------------------
WakeUpSounder::WakeUpSounder():
granInputLevel(0.2f),
distortGain(0.5f),
distortFade(0.0f),
loud(false),
glissDown(false),
waver(false),
fadeOut(1.0f),
fadeOutTweener(fadeOut, 1.0f, 0.000015f)
{
	fadeOutTweener.setIndex(1.0f);

	//scales.removeScale(L"Chromatic");
	{
		std::vector<std::wstring> scaleSet;

		//scaleSet.push_back(L"Locrian");
		scaleSet.push_back(L"Major");
		//scaleSet.push_back(L"Chinese");
		scaleSet.push_back(L"Lydian");
		//scaleSet.push_back(L"MelodicMinorAsc");
		scaleSet.push_back(L"Mixolydian");
		//scaleSet.push_back(L"MelodicMinorDesc");
		scaleSet.push_back(L"Phrygian");
		//scaleSet.push_back(L"PentatonicMinor");
		scaleSet.push_back(L"PentatonicMajor");
		//scaleSet.push_back(L"NaturalMinor");

		scales.chooseScales(scaleSet);
	}
	scales.getFrequencies(pitches,
						  48 + Random::getGlobal().getRandomInt(12),
						  2,
						  scaleName);

	blueSynth.loadParameters(BlueSound::numParameters, BlueSound::parameters);
	greenSynth.loadParameters(GreenSound::numParameters, GreenSound::parameters);
	redSynth.loadParameters(RedSound::numParameters, RedSound::parameters);
	newEntitySynth.loadParameters(NewEntity::numParameters, NewEntity::parameters);
	newTriangleSynth.loadParameters(NewTriangle::numParameters, NewTriangle::parameters);

	colourCycleFilter.setPitchSet(pitches);
	
	synthDelay.setDelay(0.25f);
	synthDelay.setFeedback(0.5f);
	synthDelay.setMix(0.33f);

	delay.setDelay(0.66f);
	delay.setFeedback(0.5f);
	delay.setMix(0.25f);
	granulators[0].setDensity(0.66f);
	granulators[0].setDuration(2.0f);
	//granulators[0].setPitch(1.0f * powf(2.0f, 5.0f/12.0f));
	granulators[0].setPitch(0.5f);
	granulators[0].setMaxBlockSize(44100*2);
	granulators[1].setDensity(1.0f);
	granulators[1].setDuration(2.0f);
	//granulators[1].setPitch(1.0f * powf(2.0f, 7.0f/12.0f));
	granulators[1].setPitch(2.0f);
	granulators[1].setMaxBlockSize(44100*2);
	bitCrusher.setBitDepth(4.0f);
	bitCrusher.setSamplerate(0.1f);
	bitCrusher.setMix(0.0f);
	combFilters.setMix(0.33f);
	combFilters.setOutputLevel(0.025f);

	chorus.setMix(1.0f);
	//flanger.setMix(1.0f);

	glissFilter[0].setCutoffAndRes(1500.0f, 0.15f);
	glissFilter[1].setCutoffAndRes(1500.0f, 0.15f);

	colourCycleFilter.setMix(0.0f);
	colourCycleFilter.setOutputLevel(0.05f);
}

//------------------------------------------------------------------------------
WakeUpSounder::~WakeUpSounder()
{

}

//------------------------------------------------------------------------------
void WakeUpSounder::getAudio(float *input, float *output, int numSamples)
{
	float dist;
	float tempf;

	assert(numSamples < (BufferSize/2));

	//Just in case.
	for(int i=0;i<(numSamples*2);++i)
		output[i] = 0.0f;

	//Generate any simple synth notes.
	blueSynth.getAudio(output, numSamples);
	greenSynth.getAudio(output, numSamples);
	redSynth.getAudio(output, numSamples);
	newEntitySynth.getAudio(output, numSamples);
	newTriangleSynth.getAudio(output, numSamples);
	synthDelay.getAudio(output, numSamples);

	//Generate sine drones.
	for(int i=0;i<NumSines;++i)
	{
		if(sines[i].getAmp() > 0.0f)
			sines[i].getAudio(output, numSamples);
	}

	//Process the granulators.
	for(int i=0;i<(numSamples*2);++i)
	{
		granBuffers[0][i] = output[i] * granInputLevel;
		granBuffers[1][i] = output[i] * granInputLevel;
	}
	granulators[0].processBlock(granBuffers[0], numSamples);
	granulators[1].processBlock(granBuffers[1], numSamples);

	//Granulators and chorus get processed in parallel.
	chorus.processAudio(output, numSamples);

	//Add the granulator output back in.
	for(int i=0;i<(numSamples*2);++i)
	{
		output[i] += granBuffers[0][i] * 2.0f;
		output[i] += granBuffers[1][i] * 1.25f;
	}

	delay.getAudio(output, numSamples);
	bitCrusher.processAudio(output, numSamples);
	flanger.processAudio(output, numSamples);
	combFilters.processBlock(output, numSamples);

	//colourCycleFilter.processBlock(output, numSamples);

	if(loud || (distortFade > 0.0f))
	{
		tempf = (1.0f - (fakeLog(distortGain) * 0.9f)) * 0.6f;
		dist = 1.0f+(fakeLog(distortGain)*16.0f);
		for(int i=0;i<numSamples;++i)
		{
			output[(i*2)] += ((tanhf(output[(i*2)]*dist) * tempf) - output[(i*2)]) * distortFade;
			output[(i*2)+1] += ((tanhf(output[(i*2)+1]*dist) * tempf) - output[(i*2)+1]) * distortFade;

			if(loud && (distortFade < 1.0f))
			{
				distortFade += 0.0001f;
				if(distortFade > 1.0f)
					distortFade = 1.0f;
			}
			else if(!loud && (distortFade > 0.0f))
			{
				distortFade -= 0.00001f;
				if(distortFade < 0.0f)
				{
					distortFade = 0.0f;
					break;
				}
			}
		}
	}

	if(glissDown)
	{
		for(int i=0;i<numSamples;++i)
		{
			output[(i*2)] = glissFilter[0].process(output[(i*2)]) * 2.0f;
			output[(i*2)+1] = glissFilter[1].process(output[(i*2)+1]) * 2.0f;
		}
	}

	for(int i=0;i<(numSamples*2);++i)
		output[i] = tanhf(output[i] * 0.9f * NIALL_PI) * 0.99f;

	if(fadeOut < 1.0f)
	{
		for(int i=0;i<numSamples;++i)
		{
			output[(i*2)] *= fadeOut;
			output[(i*2)+1] *= fadeOut;

			if(!fadeOutTweener.getFinished())
				fadeOutTweener.tick();
		}
	}
}

//------------------------------------------------------------------------------
int WakeUpSounder::startDrone(float pitch, float pan)
{
	int retval(-1);
	const float pitchVal(clamp<float>(pitch, 0.0f, 1.0f));

	for(int i=0;i<NumSines;++i)
	{
		if(sines[i].getAmp() <= 0.0f)
		{
			retval = i;

			sines[i].setPitch(pitches[(int)(pitchVal * ((float)pitches.size()-0.01f))]);
			sines[i].setPan(pan);
			sines[i].setAmp(0.1f);
			break;
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
void WakeUpSounder::stopDrone(int index)
{
	if(index > -1)
		sines[index].setAmp(0.0f);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setDroneAmp(int index, float val)
{
	if(index > -1)
		sines[index].setAmp(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setDronePan(int index, float val)
{
	if(index > -1)
		sines[index].setPan(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setDroneDist(int index, float val)
{
	if(index > -1)
		sines[index].setDist(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerBlueSynth()
{
	blueSynth.triggerNote(pitches[Random::getGlobal().getRandomInt(pitches.size())], 0);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerGreenSynth()
{
	greenSynth.triggerNote(pitches[Random::getGlobal().getRandomInt(pitches.size())], 0);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerRedSynth()
{
	redSynth.triggerNote(pitches[Random::getGlobal().getRandomInt(pitches.size())], 0);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerNewEntitySynth(float val)
{
	newEntitySynth.triggerNote(pitches[(int)(val * ((float)pitches.size() - 0.01f))], 0);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerNewTriangleSynth(float val)
{
	newTriangleSynth.triggerNote(pitches[(int)(val * ((float)pitches.size() - 0.01f))], 0);
}

//------------------------------------------------------------------------------
void WakeUpSounder::triggerFadeOut()
{
	//fadeOut -= 0.00001f;
	fadeOutTweener.reset(0.0f);
	fadeOutTweener.tick();
}

//------------------------------------------------------------------------------
void WakeUpSounder::setDelay(float val)
{
	delay.setDelay(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setDelayFeedback(float val)
{
	delay.setFeedback(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setGranInputLevel(float val)
{
	granInputLevel = val;
}

//------------------------------------------------------------------------------
void WakeUpSounder::setGranFeedback(float val)
{
	granulators[0].setFeedback(val);
	granulators[1].setFeedback(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setBitCrusherMix(float val)
{
	bitCrusher.setMix(val, true);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setChorusMix(float val)
{
	chorus.setMix(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setFlangerMix(float val)
{
	flanger.setMix(val);
}

//------------------------------------------------------------------------------
void WakeUpSounder::incDistortGain()
{
	if(distortGain < 0.5f)
		distortGain += 0.01f;
}

//------------------------------------------------------------------------------
void WakeUpSounder::setLoud(bool val)
{
	loud = val;
}

//------------------------------------------------------------------------------
void WakeUpSounder::pickNewPitches()
{
	ScopedLock lock(pitchesCritSec);
	std::wstring tempName(scaleName);

	while(tempName == scaleName)
	{
		pitches.clear();
		tempName = L"";
		scales.getFrequencies(pitches,
							  36 + Random::getGlobal().getRandomInt(24),
							  2 + Random::getGlobal().getRandomInt(2),
							  tempName);
	}
	scaleName = tempName;

	for(int i=0;(i<8)&&(i<(int)pitches.size());++i)
		combFilters.setPitch(i, pitches[i]);

	//Make sure we handle pitch sizes less than 8.
	for(int i=(int)pitches.size();i<8;++i)
		combFilters.setPitch(i, (pitches[i % pitches.size()] * 2.0f));

	colourCycleFilter.setPitchSet(pitches);
}

//------------------------------------------------------------------------------
void WakeUpSounder::setPitches(const std::vector<float>& pitchScale)
{
	ScopedLock lock(pitchesCritSec);

	pitches.clear();

	for(unsigned int i=0;i<pitchScale.size();++i)
		pitches.push_back(pitchScale[i]);

	for(int i=0;(i<8)&&(i<(int)pitches.size());++i)
		combFilters.setPitch(i, pitches[i]);

	//Make sure we handle pitch sizes less than 8.
	for(int i=(int)pitches.size();i<8;++i)
		combFilters.setPitch(i, (pitches[i % pitches.size()] * 2.0f));
}
