//	WakeUpSounder.h - The game's sounder.
//  ----------------------------------------------------------------------------
//	This file is part of 'SoManyJaggedShards', an entry for 7DRL 2013.
//	Copyright (C) 2013  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef WAKEUPSOUNDER_H_
#define WAKEUPSOUNDER_H_

#include "synth/CombFilterBankModded.h"
#include "sound/ScalesAndFrequencies.h"
#include "synth/PolyphonicSynth.h"
#include "synth/CombFilterBank.h"
#include "synth/SimpleSynth.h"
#include "synth/Granulator.h"
#include "synth/BitCrusher.h"
#include "synth/SineDrone.h"
#include "CriticalSection.h"
#include "sound/Sounder.h"
#include "synth/Flanger.h"
#include "synth/Chorus.h"
#include "sound/Delay.h"

#include <vector>

///	The game's sounder.
class WakeUpSounder : public Sounder
{
  public:
	///	Constructor.
	WakeUpSounder();
	///	Destructor.
	~WakeUpSounder();

	///	Called for every audio block.
	void getAudio(float *input, float *output, int numSamples);

	///	Called for each sound file resource by LoadingScreen.
	void loadSoundFileResource(SoundFileLoaderPtr soundFileLoader) {};

	///	Starts a sine drone.
	/*!
		\param pitch The pitch of the drone (0->1) (taken from the current
		scale).
		\param pan How the drone should be panned.
		\return The index of the drone created. Will return -1 if all drones are
		in use.
	 */
	int startDrone(float pitch, float pan);
	///	Stops the indexed sine drone.
	void stopDrone(int index);
	///	Sets the indexed sine drone's amplitude.
	void setDroneAmp(int index, float val);
	///	Sets the indexed sine drone's panning.
	void setDronePan(int index, float val);
	///	Sets the indexed sine drone's distortion.
	void setDroneDist(int index, float val);

	///	Triggers the blue synth.
	void triggerBlueSynth();
	///	Triggers the green synth.
	void triggerGreenSynth();
	///	Triggers the red synth.
	void triggerRedSynth();
	///	Triggers the new entity synth.
	void triggerNewEntitySynth(float val);
	///	Triggers the new triangle synth.
	void triggerNewTriangleSynth(float val);

	///	Triggers a fade out.
	void triggerFadeOut();

	///	Sets the delay time (in seconds).
	void setDelay(float val);
	///	Sets the delay feedback.
	void setDelayFeedback(float val);

	///	Sets the granulator input level.
	void setGranInputLevel(float val);
	///	Sets the granulator feedback.
	void setGranFeedback(float val);

	///	Sets the bit crusher's mix.
	void setBitCrusherMix(float val);

	///	Sets the chorus' mix.
	void setChorusMix(float val);
	///	Sets the flanger's mix.
	void setFlangerMix(float val);

	///	Increments the distortion gain.
	void incDistortGain();
	///	Returns the current distortion gain.
	float getDistortGain() const {return distortGain;};
	///	Turns the distortion on/off.
	void setLoud(bool val);
	///	Returns whether the distortion is on or off.
	bool getLoud() const {return loud;};

	///	Picks a new random pitch set (that isn't the last one).
	void pickNewPitches();
	///	Sets the pitch set to use.
	void setPitches(const std::vector<float>& pitchScale);
	///	Returns the name of the current scale.
	const std::wstring& getScaleName() const {return scaleName;};
	///	Returns the current scale's pitches.
	const std::vector<float>& getPitches() const {return pitches;};
  private:
	///	Some constants.
	enum
	{
		BufferSize = 44100*2*2,
		NumSines = 128
	};

	///	Our sine drones.
	SineDrone sines[NumSines];

	///	Our SimpleSynths.
	PolyphonicSynth<SimpleSynth> blueSynth;
	///	Our SimpleSynths.
	PolyphonicSynth<SimpleSynth> greenSynth;
	///	Our SimpleSynths.
	PolyphonicSynth<SimpleSynth> redSynth;
	///	Our SimpleSynths.
	PolyphonicSynth<SimpleSynth> newEntitySynth;
	///	Our SimpleSynths.
	PolyphonicSynth<SimpleSynth> newTriangleSynth;
	///	The simple synth delay.
	Delay synthDelay;

	///	The delay on the output.
	Delay delay;
	///	The 2 granulators.
	Granulator granulators[2];
	///	The chorus on the output.
	Chorus chorus;
	///	The bit crusher.
	BitCrusher bitCrusher;
	///	The (kaleidoscope) flanger.
	Flanger flanger;
	///	The comb filter bank.
	CombFilterBank combFilters;
	///	The dither granulator.
	Granulator ditherGran;
	///	The colour-cycle glissando-ing comb filter.
	CombFilterBankModded colourCycleFilter;

	///	Input level for the granulators.
	float granInputLevel;
	///	Buffers for the 2 granulators.
	float granBuffers[2][BufferSize];

	///	Used to distort the output.
	float distortGain;
	///	Used to fade in/out the distortion.
	float distortFade;
	///	Whether to distort the output or not.
	bool loud;

	///	True if we're currently an octave lower.
	bool glissDown;
	///	Filter for the gliss down sound.
	MoogFilter glissFilter[2];

	///	True if we're currently doing a pitch waver.
	bool waver;

	///	Handles scale/frequency selection.
	ScalesAndFrequencies scales;
	///	The current scale.
	std::vector<float> pitches;
	///	The name of the current scale.
	std::wstring scaleName;
	///	CriticalSection we use to protect pitches.
	CriticalSection pitchesCritSec;

	///	Used to fade out when the game is quit.
	float fadeOut;
	///	Used to fade out.
	LogTweener<float> fadeOutTweener;
};

#endif
