//	ParameterRandomiser.cpp - Class used to randomise synth parameters
//							  'meaningfully'.
//  ----------------------------------------------------------------------------
//	This file is part of 'Hummingbird', an audiovisual ecosystem thing.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "ParameterRandomiser.h"
#include "DebugHeaders.h"
#include "Random.h"

#include <cassert>

//------------------------------------------------------------------------------
template<typename T>
const std::map<std::wstring, T>& ParameterRandomiser<T>::getRandomValues()
{
	returnMap.clear();

	if(dependencyTree.size() == 0)
		populateDependencyTree();

	//First we generate the normalised values for all the params.
	for(unsigned int i=0;i<dependencyTree.size();++i)
		returnMap.insert(std::make_pair(dependencyTree[i], random()));

	//Now we convert the normalised values to the correct ranges.
	//(starting with the dependant params)
	for(int i=((int)dependencyTree.size()-1);i>=0;--i)
	{
		//Convert dependant parameters to the correct range.
		if(parameters[dependencyTree[i]].dependency > -1)
		{
			Dependency<T>& d(dependencies[parameters[dependencyTree[i]].dependency]);

			returnMap[dependencyTree[i]] = d.convertToRange(returnMap[dependencyTree[i]],
															returnMap[d.sourceParam]);
		}
		//And convert non-dependant parameters to the correct range.
		else
		{
			Parameter<T>& p(parameters[dependencyTree[i]]);

			returnMap[dependencyTree[i]] = p.convertToRange(returnMap[dependencyTree[i]]);
		}
	}

	return returnMap;
}

//------------------------------------------------------------------------------
template<typename T>
void ParameterRandomiser<T>::registerParameter(const std::wstring& name,
											   const T& rangeMin,
											   const T& rangeMax,
											   CurveFunction *curveFunc)
{
	Parameter<T> param;

	param.rangeMin = rangeMin;
	param.rangeMax = rangeMax;
	param.curveFunction = curveFunc;

	parameters.insert(std::make_pair(name, param));
	dependencyTree.clear();
}

//------------------------------------------------------------------------------
template<typename T>
void ParameterRandomiser<T>::setParameter(const std::wstring& name,
										  const T& rangeMin,
										  const T& rangeMax,
										  CurveFunction *curveFunc)
{
	parameters[name].rangeMin = rangeMin;
	parameters[name].rangeMax = rangeMax;
	parameters[name].curveFunction = curveFunc;

	dependencyTree.clear();
}

//------------------------------------------------------------------------------
template<typename T>
void ParameterRandomiser<T>::registerDependency(const std::wstring& sourceParam,
												const std::wstring& dependantParam,
												const T& minRangeMin,
												const T& minRangeMax,
												CurveFunction *minRangeCurve,
												const T& maxRangeMin,
												const T& maxRangeMax,
												CurveFunction *maxRangeCurve,
												CurveFunction *interpCurve)
{
	if(dependencyIsValid(sourceParam, dependantParam))
	{
		Dependency<T> dependency;

		dependency.sourceParam = sourceParam;
		dependency.dependantParam = dependantParam;

		dependency.minRangeMin = minRangeMin;
		dependency.minRangeMax = minRangeMax;
		dependency.minRangeCurve = minRangeCurve;

		dependency.maxRangeMin = maxRangeMin;
		dependency.maxRangeMax = maxRangeMax;
		dependency.maxRangeCurve = maxRangeCurve;

		dependency.interpCurve = interpCurve;

		dependencies.push_back(dependency);
		dependencyTree.clear();

		parameters[dependantParam].dependency = dependencies.size()-1;
	}
	else
		assert(true);
}

//------------------------------------------------------------------------------
template<typename T>
void ParameterRandomiser<T>::populateDependencyTree()
{
	std::vector<std::wstring> remainingParams;
	typename std::map<std::wstring, Parameter<T> >::iterator it;

	//Stick all our parameters in a vector.
	for(it=parameters.begin();it!=parameters.end();++it)
		remainingParams.push_back(it->first);

	//Move any parameters with no dependencies into our dependency tree.
	for(int i=((int)remainingParams.size()-1);i>=0;--i)
	{
		if(parameters[remainingParams[i]].dependency < 0)
		{
			dependencyTree.push_back(remainingParams[i]);

			remainingParams.erase(remainingParams.begin()+i);
		}
	}

	//Now go through any remaining remainingParams...
	while(remainingParams.size() > 0)
	{
		//...and move any whose source param has already been moved into the dependency tree.
		for(int i=((int)remainingParams.size()-1);i>=0;--i)
		{
			const Dependency<T>& dependency(dependencies[parameters[remainingParams[i]].dependency]);
			const std::wstring& source(dependency.sourceParam);
			bool foundSource(false);

			for(unsigned int j=0;j<remainingParams.size();++j)
			{
				if(remainingParams[j] == source)
				{
					foundSource = true;
					break;
				}
			}

			if(!foundSource)
			{
				dependencyTree.push_back(remainingParams[i]);

				remainingParams.erase(remainingParams.begin()+i);
			}
		}
	}
}

//------------------------------------------------------------------------------
template<typename T>
bool ParameterRandomiser<T>::dependencyIsValid(const std::wstring& source,
											   const std::wstring& dependant)
{
	bool retval(true);

	for(unsigned int i=0;i<dependencies.size();++i)
	{
		if(dependencies[i].dependantParam == dependant)
		{
			DebugStatement(L"ParameterRandomiser: Cannot create Dependency; dependant param is already dependant on another param.");
			DebugStatement(L"Source param = " << source << L" : Dependant param = " << dependant);
			retval = false;
			break;
		}
		else if((dependencies[i].dependantParam == source) &&
				(dependencies[i].sourceParam == dependant))
		{
			DebugStatement(L"ParameterRandomiser: Cannot create Dependency; circular dependancy.");
			DebugStatement(L"Source param = " << source << L" : Dependant param = " << dependant);
			retval = false;
			break;
		}
	}

	return retval;
}

//------------------------------------------------------------------------------
template<> float ParameterRandomiser<float>::random() {return Random::getGlobal().getRandomFloat();};

//------------------------------------------------------------------------------
template class ParameterRandomiser<float>;
