//	ParameterRandomiser.h - Class used to randomise synth parameters
//							'meaningfully'.
//  ----------------------------------------------------------------------------
//	This file is part of 'Hummingbird', an audiovisual ecosystem thing.
//	Copyright (C) 2014  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#ifndef PARAMETERRANDOMISER_H_
#define PARAMETERRANDOMISER_H_
#include <vector>
#include <string>
#include <map>

///	Class used to randomise synth parameters 'meaningfully'.
template <typename T>
class ParameterRandomiser
{
  public:
	///	Returns a map of randomised values for all the registered parameters.
	/*!
		\return The map's format is the parameter name as the key, its
		randomised value as the value.
	 */
	const std::map<std::wstring, T>& getRandomValues();

	///	Typedef for the parameter curve function pointer.
	typedef T CurveFunction(T);

	///	Registers a parameter and its range.
	/*!
		\param name The name of the parameter.
		\param rangeMin The minimum value of the range the parameter may be
		randomised within.
		\param rangeMax The maximum value of the range the parameter may be
		randomised within.
		\param curveFunc A function which defines how the randomised values are
		distributed within the range. If this is zero (null) a linear curve is
		used.
	 */
	void registerParameter(const std::wstring& name,
						   const T& rangeMin,
						   const T& rangeMax,
						   CurveFunction *curveFunc = 0);
	///	Sets the named parameter's variables.
	void setParameter(const std::wstring& name,
					  const T& rangeMin,
					  const T& rangeMax,
					  CurveFunction *curveFunc = 0);

	///	Registers a dependency between parameters.
	/*!
		\param sourceParam The name of the parameter whose value will determine
		how the dependant parameter's range is calculated.
		\param dependantParam The name of the parameter whose range is dependant
		on the source parameter's randomised value.
		\param minRangeMin The minimum value of the dependant parameter's
		minimum range.
		\param minRangeMax The maximum value of the dependant parameter's
		minimum range.
		\param minRangeCurve The curve function of the dependant parameter's
		minimum range (0 == linear).
		\param maxRangeMin The minimum value of the dependant parameter's
		maximum range.
		\param maxRangeMax The maximum value of the dependant parameter's
		maximum range.
		\param maxRangeCurve The curve function of the dependant parameter's
		maximum range (0 == linear).
		\param interpCurve Curve function used when interpolating between the
		minimum and maximum range values.

		The dependant parameter's randomised value will be calculated by first
		calculating it's position within the parameter's two ranges (min & max),
		then interpolating between the two values according to the source
		parameter's randomised value.
	 */
	void registerDependency(const std::wstring& sourceParam,
							const std::wstring& dependantParam,
							const T& minRangeMin,
							const T& minRangeMax,
							CurveFunction *minRangeCurve,
							const T& maxRangeMin,
							const T& maxRangeMax,
							CurveFunction *maxRangeCurve = 0,
							CurveFunction *interpCurve = 0);
  private:
	///	Helper method. Fills out the dependency 'tree'.
	void populateDependencyTree();

	///	Helper method. Checks whether a Dependency between the two passed-in parameters would be valid.
	bool dependencyIsValid(const std::wstring& source,
						   const std::wstring& dependant);

	///	Helper method. Returns a random value.
	T random() {return T();};

	///	Represents a Parameter.
	template <typename T2>
	struct Parameter
	{
		///	The minimum value of the parameter's randomisation range.
		T2 rangeMin;
		///	The maximum value of the parameter's randomisation range.
		T2 rangeMax;
		///	The curve function determining how values are distributed within the randomisation range.
		CurveFunction *curveFunction;

		///	Index of this Parameter's Dependency.
		/*!
			-1 if it has no Dependency.
		 */
		int dependency;

		///	Default constructor.
		Parameter():
		curveFunction(0),
		dependency(-1)
		{

		};

		///	Converts the passed-in value to the parameter's range.
		/*!
			\param value This should be in the range 0->1.
		 */
		T2 convertToRange(const T2& value) const
		{
			T retval(value);

			retval *= (rangeMax - rangeMin);
			retval += rangeMin;

			if(curveFunction)
				retval = curveFunction(value);

			return retval;
		};
	};

	///	Represents a dependency between two parameters.
	template <typename T2>
	struct Dependency
	{
		///	The source parameter.
		std::wstring sourceParam;
		///	The dependant parameter.
		std::wstring dependantParam;

		///	The minimum value of the minimum randomisation range.
		T2 minRangeMin;
		///	The maximum value of the minimum randomisation range.
		T2 minRangeMax;
		///	The curve function for the minimum randomisation range.
		CurveFunction *minRangeCurve;

		///	The miimum value of the maximum randomisation range.
		T maxRangeMin;
		///	The maximum value of the maximum randomisation range.
		T maxRangeMax;
		///	The curve function for the maximum randomisation range.
		CurveFunction *maxRangeCurve;

		///	The curve used to interpolate between the two ranges.
		CurveFunction *interpCurve;

		///	Constructor.
		Dependency():
		minRangeCurve(0),
		maxRangeCurve(0),
		interpCurve(0)
		{

		};

		///	Converts the passed-in value to the dependant parameter's interpolated range.
		/*!
			\param value A value 0->1.
			\param sourceValue A value 0->1. Used to interpolate between the
			dependant parameter's two randomisation ranges.
		 */
		T2 convertToRange(const T2& value, const T2& sourceValue) const
		{
			T2 retval(sourceValue);
			T2 minRangeVal(value);
			T2 maxRangeVal(value);

			minRangeVal *= (minRangeMax - minRangeMin);
			minRangeVal += minRangeMin;
			if(minRangeCurve)
				minRangeVal = minRangeCurve(minRangeVal);

			maxRangeVal *= (maxRangeMax - maxRangeMin);
			maxRangeVal += maxRangeMin;
			if(maxRangeCurve)
				maxRangeVal = maxRangeCurve(maxRangeVal);

			if(interpCurve)
				retval = interpCurve(retval);
			retval = minRangeVal + ((maxRangeVal - minRangeVal) * retval);

			return retval;
		};
	};

	///	Our map of Parameters.
	std::map<std::wstring, Parameter<T> > parameters;
	///	Our vector of Dependencies.
	std::vector<Dependency<T> > dependencies;

	///	Vector holding the parameter names in the order they need to be calculated in order to fulfil the various dependencies.
	std::vector<std::wstring> dependencyTree;

	///	The map we return from getRandomValues().
	std::map<std::wstring, T> returnMap;
};

///	Float specialisation of the private random() method.
//template<> float ParameterRandomiser<float>::random() {return Random::getGlobal().getRandomFloat();};

#endif
