//	SimpleSynth.cpp - Simple analogue-esque synth.
//  ----------------------------------------------------------------------------
//	This file is part of 'GameTemplate', template code for a game.
//	Copyright (C) 2010  Niall Moody
//	
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "SimpleSynth.h"
#include "MiscHelper.h"
#include "Random.h"

#include <cmath>

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
SimpleSynth::SimpleSynth():
MonophonicSynth(),
waveAmount(0.0f),
detunedSawAmount(0.0f),
filterCutoff(0.5f),
filterRes(0.0f),
filterAttackCurve(false),
filterReleaseCurve(false),
filterAttackTime(0.25f),
filterReleaseTime(0.5f),
filterEnvAmount(0.01f),
attackCurve(false),
releaseCurve(false),
attackTime(0.25f),
releaseTime(0.5f),
noteLength(0.5f),
distGain(1.0f),
octave(1.0f),
level(1.0f),
glissDown(1.0f),
waver(false),
morphTweener(morphParams, morphParams, 0.0001f),
noteAmp(1.0f),
tableIndex(0.0f),
detuneIndex(0.0f),
frequency(440.0f),
detuneFreq(439.0f),
glissDownTween(glissDown, glissDown, 0.00002f),
waverPitch(1.0f),
waverTween(waverPitch, waverPitch, 0.000005f + (Random::getGlobal().getRandomFloat() * 0.0001f)),
noteIndex(0.0f),
increment(1.0f/44100.0f),
noteWaiting(false),
waitingSample(0),
ignoreLength(false)
{
	tableSize = WavetableSingleton::getInstance().getTableSize();
	/*sawTable = WavetableSingleton::getInstance().getSawTable();
	squareTable = WavetableSingleton::getInstance().getSquareTable();*/

	waverTween.setIndex(1.0f);

	morphTweener.setIndex(1.0f);
}

//------------------------------------------------------------------------------
SimpleSynth::~SimpleSynth()
{
	
}

//------------------------------------------------------------------------------
void SimpleSynth::getAudio(float *buffer, int numSamples)
{
	int i;
	float out;
	float saw;
	float saw2;
	float tempf;
	int tempint;
	float square;
	float ampEnv;
	float filtEnv;
	const float invSampleRate(1.0f/44100.0f);
	const float tableSizeFloat((float)tableSize);
	const float actualNoteLength = noteLength * 2.0f;
	const float attackLength = attackTime * (actualNoteLength*0.999f);
	const float releaseLength = releaseTime * (actualNoteLength-attackLength);
	const float holdEnd = attackLength + (actualNoteLength - (attackLength+releaseLength));
	const float filtAttack = filterAttackTime * (actualNoteLength*0.999f);
	const float filtRelease = filterReleaseTime * (actualNoteLength-filtAttack);
	const float filtHold = filtAttack + (actualNoteLength - (filtAttack+filtRelease));

	if(!getActive())
		return;

	for(i=0;i<numSamples;++i)
	{
		if(noteWaiting && (waitingSample >= 0))
		{
			--waitingSample;

			if(waitingSample <= 0)
				noteWaiting = false;
			else
				continue;
		}
		//Increment our position within the note, quit if we've gone past
		//noteLength.
		if(!ignoreLength || (noteIndex < holdEnd))
		{
			noteIndex += increment;
			if(noteIndex > actualNoteLength)
			{
				setActive(false);
				break;
			}
		}

		//Update the parameter morphing if necessary.
		if(!morphTweener.getFinished())
		{
			/*const std::wstring names[] = {L"waveAmount",
										  L"detunedSawAmount",
										  L"filterCutoff",
										  L"filterRes",
										  L"filterAttackCurve",
										  L"filterReleaseCurve",
										  L"filterAttackTime",
										  L"filterReleaseTime",
										  L"filterEnvAmount",
										  L"attackCurve",
										  L"releaseCurve",
										  L"attackTime",
										  L"releaseTime",
										  L"noteLength",
										  L"distGain",
										  L"octave",
										  L"level"};*/

			morphTweener.tick();

			for(int j=0;j<17;++j)
				setParameter(j, morphParams[j]);
		}

		//Calculate the amplitude envelope.
		if(noteIndex < attackLength)
		{
			if(attackCurve)
				ampEnv =  fakeExp(noteIndex/attackLength);
			else
				ampEnv =  fakeLog(noteIndex/attackLength);
		}
		else if(noteIndex < holdEnd)
			ampEnv = 1.0f;
		else
		{
			if(releaseCurve)
				ampEnv = 1.0f-fakeLog((noteIndex-holdEnd)/releaseLength);
			else
				ampEnv = 1.0f-fakeExp((noteIndex-holdEnd)/releaseLength);
		}

		//Calculate the filter envelope.
		if(noteIndex < filtAttack)
		{
			if(filterAttackCurve)
				filtEnv =  fakeExp(noteIndex/filtAttack) * filterEnvAmount;
			else
				filtEnv =  fakeLog(noteIndex/filtAttack) * filterEnvAmount;
		}
		else if(noteIndex < filtHold)
			filtEnv = filterEnvAmount;
		else
		{
			if(filterReleaseCurve)
				filtEnv = (1.0f-fakeLog((noteIndex-filtHold)/filtRelease)) * filterEnvAmount;
			else
				filtEnv = (1.0f-fakeExp((noteIndex-filtHold)/filtRelease)) * filterEnvAmount;
		}

		//Generate the sawtooth audio.
		tempint = (int)tableIndex;
		tempf = tableIndex - (float)tempint;
		saw = sawTable[tempint] * (1.0f-tempf);
		if(tempint < (tableSize-1))
			saw += sawTable[tempint+1] * tempf;
		else
			saw += sawTable[0] * tempf;

		//Generate the square wave audio.
		tempint = (int)tableIndex;
		tempf = tableIndex - (float)tempint;
		square = squareTable[tempint] * (1.0f-tempf);
		if(tempint < (tableSize-1))
			square += squareTable[tempint+1] * tempf;
		else
			square += squareTable[0] * tempf;

		//Mix the 2 waveforms together.
		out = saw * (1.0f-waveAmount);
		out += square * waveAmount;

		//Add the detuned sawtooth.
		if(detunedSawAmount > 0.0f)
		{
			tempint = (int)detuneIndex;
			tempf = detuneIndex - (float)tempint;
			saw2 = sawTable[tempint] * (1.0f-tempf);
			if(tempint < (tableSize-1))
				saw2 += sawTable[tempint+1] * tempf;
			else
				saw2 += sawTable[0] * tempf;

			out += saw2 * detunedSawAmount;
		}

		//Handle any glissando down.
		if(!glissDownTween.getFinished())
			glissDownTween.tick();

		//Work out the wavering.
		if(!waverTween.getFinished())
		{
			waverTween.tick();

			if(waverTween.getFinished())
			{
				if(waver)
				{
					/*if(waverPitch > 1.0f)
						waverTween.reset(1.0f);
					else
						waverTween.reset(2.0f);*/
					waverTween.reset(1.0f + ((Random::getGlobal().getRandomFloat() * 0.1f)-0.05f));
				}
			}
		}

		//Update the table index.
		//tableIndex += (float)tableSize/(44100.0f/(frequency * glissDown * waverPitch));
		tableIndex += (tableSizeFloat * (frequency * glissDown * waverPitch)) * invSampleRate;
		if(tableIndex > tableSizeFloat)
			tableIndex -= tableSizeFloat;

		//Update the detuned sawtooth index.
		//detuneIndex += (float)tableSize/(44100.0f/(detuneFreq * glissDown));
		detuneIndex += (tableSizeFloat * (detuneFreq * glissDown)) * invSampleRate;
		if(detuneIndex > tableSizeFloat)
			detuneIndex -= tableSizeFloat;

		//Update the filter's parameters.
		tempf = filterCutoff * (11025.0f-50.0f);
		if(filtEnv < 0.0f)
			tempf += filtEnv * tempf;
		else
			tempf += filtEnv * (11025.0f-tempf);
		filter.setCutoffAndRes(50.0f + tempf, filterRes);

		//Apply the filter to the waves.
		out = filter.process(out);

		//Apply the distortion.
		out = tanhf(out * distGain);

		//Apply the amplitude envelope to the output.
		out *= ampEnv;
		out *= 0.25f * level * noteAmp;

		buffer[i*2] += out;
		buffer[(i*2)+1] += out;
	}
}

//------------------------------------------------------------------------------
void SimpleSynth::triggerNote(float freq, int sample, float amp)
{
	frequency = freq * octave;
	detuneFreq = frequency * 0.999f;
	glissDown = glissDownTween.getEndVal();
	glissDownTween.setIndex(1.0f);

	sawTable = WavetableSingleton::getInstance().getSawTable(frequency);
	squareTable = WavetableSingleton::getInstance().getSquareTable(frequency);

	noteWaiting = true;
	waitingSample = sample;

	noteAmp = amp;

	//tableIndex = 0.0f;
	noteIndex = 0.0f;
	setActive(true);
}

//------------------------------------------------------------------------------
void SimpleSynth::noteOn(float freq, int sample)
{
	frequency = freq * octave;
	detuneFreq = frequency * 0.999f;
	glissDown = glissDownTween.getEndVal();
	glissDownTween.setIndex(1.0f);

	sawTable = WavetableSingleton::getInstance().getSawTable(frequency);
	squareTable = WavetableSingleton::getInstance().getSquareTable(frequency);

	noteWaiting = true;
	waitingSample = sample;

	//tableIndex = 0.0f;
	noteIndex = 0.0f;
	setActive(true);

	ignoreLength = true;
}

//------------------------------------------------------------------------------
void SimpleSynth::noteOff(int sample)
{
	ignoreLength = false;
}

//------------------------------------------------------------------------------
void SimpleSynth::setParameter(const std::wstring& name, float value, bool direct)
{
	if(name == L"waveAmount")
		waveAmount = value;
	else if(name == L"detunedSawAmount")
		detunedSawAmount = value;
	else if(name == L"filterCutoff")
	{
		if(direct)
			filterCutoff = value;
		else
			filterCutoff = (fakeExp(value) * 0.5f) + (value * 0.5f);
	}
	else if(name == L"filterRes")
		filterRes = value;
	else if(name == L"filterAttackCurve")
		filterAttackCurve = (value > 0.5f);
	else if(name == L"filterReleaseCurve")
		filterReleaseCurve = (value > 0.5f);
	else if(name == L"filterAttackTime")
		filterAttackTime = value;
	else if(name == L"filterReleaseTime")
		filterReleaseTime = value;
	else if(name == L"filterEnvAmount")
		filterEnvAmount = value;
	else if(name == L"attackCurve")
		attackCurve = (value > 0.5f);
	else if(name == L"releaseCurve")
		releaseCurve = (value > 0.5f);
	else if(name == L"attackTime")
		attackTime = value;
	else if(name == L"releaseTime")
		releaseTime = value;
	else if(name == L"noteLength")
	{
		if(direct)
			noteLength = value;
		else
			noteLength = 0.05f + (value*0.95f);
	}
	else if(name == L"distGain")
	{
		if(direct)
			distGain = value;
		else
			distGain = 1.0f + (value * 3.0f);
	}
	else if(name == L"octave")
	{
		if(direct)
			octave = value;
		else
		{
			if(value < 0.9f)
				octave = 4.0f;
			else if(value < 1.9f)
				octave = 2.0f;
			else if(value < 2.9f)
				octave = 1.0f;
			else if(value < 3.9f)
				octave = 0.5f;
			else
				octave = 0.25f;
		}
	}
	else if(name == L"level")
		level = value;
	else if(name == L"glissDown")
	{
		//if(glissDownTween.getFinished())
		{
			//if(glissDownTween.getEndVal() > 0.5f)
			if(value > 0.5f)
				glissDownTween.reset(0.5f);
			else
				glissDownTween.reset(1.0f);
		}
	}
	else if(name == L"waver")
	{
		if(value < 0.5f)
			waver = false;
		else
			waver = true;

		if(!waver)
		{
			waverTween.reset(1.0f);
			if(!getActive())
			{
				waverPitch = 1.0f;
				waverTween.setIndex(1.0f);
			}
		}
		else
			waverTween.reset(1.0f + ((Random::getGlobal().getRandomFloat() * 0.066f)-0.033f));
	}
}

//------------------------------------------------------------------------------
void SimpleSynth::setParameter(int index, float value)
{
	switch(index)
	{
		case 0:
			waveAmount = value;
			break;
		case 1:
			filterCutoff = (fakeExp(value) * 0.5f) + (value * 0.5f);
			break;
		case 2:
			filterRes = value;
			break;
		case 3:
			filterAttackCurve = (value > 0.5f);
			break;
		case 4:
			filterReleaseCurve = (value > 0.5f);
			break;
		case 5:
			filterAttackTime = value;
			break;
		case 6:
			filterReleaseTime = value;
			break;
		case 7:
			filterEnvAmount = value;
			break;
		case 8:
			attackCurve = (value > 0.5f);
			break;
		case 9:
			releaseCurve = (value > 0.5f);
			break;
		case 10:
			attackTime = value;
			break;
		case 11:
			releaseTime = value;
			break;
		case 12:
			noteLength = 0.05f + (value*0.95f);
			break;
		case 13:
				distGain = 1.0f + (value * 3.0f);
				break;
		case 14:
			if(value < 0.9f)
				octave = 4.0f;
			else if(value < 1.9f)
				octave = 2.0f;
			else if(value < 2.9f)
				octave = 1.0f;
			else if(value < 3.9f)
				octave = 0.5f;
			else
				octave = 0.25f;
			break;
		case 15:
			detunedSawAmount = value;
			break;
		case 16:
			level = value;
			break;
		case 17:
			if(value > 0.5f)
					glissDownTween.reset(0.5f);
				else
					glissDownTween.reset(1.0f);
			break;
		case 18:
			if(value < 0.5f)
				waver = false;
			else
				waver = true;

			if(!waver)
			{
				waverTween.reset(1.0f);
				if(!getActive())
				{
					waverPitch = 1.0f;
					waverTween.setIndex(1.0f);
				}
			}
			else
				waverTween.reset(1.0f + ((Random::getGlobal().getRandomFloat() * 0.066f)-0.033f));
			break;
	}
}

//------------------------------------------------------------------------------
void SimpleSynth::setAllParameters(const std::map<std::wstring, float>& values,
								   float morphTime)
{
	if((morphTime <= 0.0f) || !getActive())
	{
		std::map<std::wstring, float>::const_iterator it;

		for(it=values.begin();it!=values.end();++it)
			setParameter(it->first, it->second, false);
	}
	else
	{
		NiallArray<float, 17> inputArr;

		//Make sure our morph params match our actual params.
		setMorphParams();

		morphTweener.setDelta(morphTime/44100.0f);
		convertMapToArray(values, inputArr);
		morphTweener.reset(inputArr);
	}
}

//------------------------------------------------------------------------------
float SimpleSynth::getParameter(int i) const
{
	switch(i)
	{
		case 0:
			return waveAmount;
		case 1:
			return filterCutoff;
		case 2:
			return filterRes;
		case 3:
			return filterAttackCurve;
		case 4:
			return filterReleaseCurve;
		case 5:
			return filterAttackTime;
		case 6:
			return filterReleaseTime;
		case 7:
			return filterEnvAmount;
		case 8:
			return attackCurve;
		case 9:
			return releaseCurve;
		case 10:
			return attackTime;
		case 11:
			return releaseTime;
		case 12:
			return noteLength;
		case 13:
			return distGain;
		case 14:
			return octave;
		case 15:
			return detunedSawAmount;
		case 16:
			return level;
		case 17:
			return glissDown;
		default:
			return 0.0f;
	}
}

//------------------------------------------------------------------------------
const std::wstring SimpleSynth::getParameterName(int i) const
{
	switch(i)
	{
		case 0:
			return L"waveAmount";
		case 1:
			return L"filterCutoff";
		case 2:
			return L"filterRes";
		case 3:
			return L"filterAttackCurve";
		case 4:
			return L"filterReleaseCurve";
		case 5:
			return L"filterAttackTime";
		case 6:
			return L"filterReleaseTime";
		case 7:
			return L"filterEnvAmount";
		case 8:
			return L"attackCurve";
		case 9:
			return L"releaseCurve";
		case 10:
			return L"attackTime";
		case 11:
			return L"releaseTime";
		case 12:
			return L"noteLength";
		case 13:
			return L"distGain";
		case 14:
			return L"octave";
		case 15:
			return L"detunedSawAmount";
		case 16:
			return L"level";
		case 17:
			return L"glissDown";
		case 18:
			return L"waver";
		default:
			return L"";
	}
}

//------------------------------------------------------------------------------
void SimpleSynth::setMorphParams()
{
	morphParams[0] = waveAmount;
	morphParams[1] = detunedSawAmount;
	morphParams[2] = filterCutoff;
	morphParams[3] = filterRes;
	morphParams[4] = filterAttackCurve;
	morphParams[5] = filterReleaseCurve;
	morphParams[6] = filterAttackTime;
	morphParams[7] = filterReleaseTime;
	morphParams[8] = filterEnvAmount;
	morphParams[9] = attackCurve;
	morphParams[10] = releaseCurve;
	morphParams[11] = attackTime;
	morphParams[12] = releaseTime;
	morphParams[13] = noteLength;
	morphParams[14] = distGain;
	morphParams[15] = octave;
	morphParams[16] = level;
}

//------------------------------------------------------------------------------
void SimpleSynth::convertMapToArray(std::map<std::wstring, float> params,
								    NiallArray<float, 17>& arr)
{
	const std::wstring names[] = {L"waveAmount",
								  L"detunedSawAmount",
								  L"filterCutoff",
								  L"filterRes",
								  L"filterAttackCurve",
								  L"filterReleaseCurve",
								  L"filterAttackTime",
								  L"filterReleaseTime",
								  L"filterEnvAmount",
								  L"attackCurve",
								  L"releaseCurve",
								  L"attackTime",
								  L"releaseTime",
								  L"noteLength",
								  L"distGain",
								  L"octave",
								  L"level"};

	for(int i=0;i<17;++i)
	{
		std::map<std::wstring, float>::iterator it(params.find(names[i]));

		if(it != params.end())
			arr[i] = it->second;
		else
			arr[i] = 0.0f;
	}
}

//------------------------------------------------------------------------------
SimpleSynth::WavetableSingleton& SimpleSynth::WavetableSingleton::getInstance()
{
	static WavetableSingleton retval;

	return retval;
}
